/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * dia-render-script.c - plugin for dia
 * Copyright (C) 2009, Hans Breuer, <Hans@Breuer.Org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Dia Render File - Serialization of Dia  Renderers.
 * May also provide some DiaObject implementation which could offer parameters like
 * - scale: overall scaling of the 'object diagram' done on the renderer level
 * - tweaking renderer parameters otherwise not in the UI, e.g. LineCaps
 *
 * The format is similar to Windows MetaFile.i.e a serialization of function calls together 
 * with their parameters.
 * An XML representation as well as some binary variant could be provided. 
 *
 * <diagram>
 *   <meta />
 *   <layer name="" >
 *     <object type="">
 *       <!-- optional stdprops to restore the real object type -->
 *       <properties />
 *       <render>
 *          <set-font />
 *          <draw-rectangle />
 *          <fill-ellipse />
 *          <draw-string />
 *          <!-- ... -->
 *       </render>
 *     </object>
 *   </layer>
 * </diagram>
 */
 
#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define G_LOG_DOMAIN "DiaRenderScript"
#include <glib.h>
#include <glib/gstdio.h>

#include "intl.h"
#include "filter.h"
#include "plug-ins.h"
#include "message.h"
#include "diagramdata.h"
#include "dia_xml_libxml.h"

#include "dia-render-script-renderer.h"

static void
drs_render_layer (DiaRenderer *self, Layer *layer)
{
  DrsRenderer *renderer = DRS_RENDERER (self);
  xmlNodePtr node;
  GList *list;
  DiaObject *obj;
  DiaRendererClass *renderer_class = DIA_RENDERER_GET_CLASS(renderer);

  g_queue_push_tail (renderer->parents, renderer->root);
  renderer->root = node = xmlNewChild(renderer->root, NULL, (const xmlChar *)"layer", NULL);
  xmlSetProp(node, (const xmlChar *)"name", (xmlChar *)layer->name);
  xmlSetProp(node, (const xmlChar *)"visible", (xmlChar *)(layer->visible ? "true" : "false"));

  /* Draw all objects: */
  list = layer->objects;
  while (list!=NULL) {
    obj = (DiaObject *) list->data;
    renderer_class->draw_object(self, obj);
    list = g_list_next(list);
  }

  renderer->root = g_queue_pop_tail (renderer->parents);
}

/*! own version to render invisible layers, too */
static void
drs_data_render (DiagramData *data, DiaRenderer *renderer)
{
  int i;
  
  DIA_RENDERER_GET_CLASS(renderer)->begin_render(renderer);
  for (i=0; i<data->layers->len; i++) {
    Layer *layer = (Layer *) g_ptr_array_index(data->layers, i);
    drs_render_layer (renderer, layer);
  }
  DIA_RENDERER_GET_CLASS(renderer)->end_render(renderer);
}

/* dia export funtion */
static void
export_data(DiagramData *data, const gchar *filename, 
            const gchar *diafilename, void* user_data)
{
  DrsRenderer *renderer;
  xmlDtdPtr dtd;
  xmlDocPtr doc;
  
  /* write check - almost same code in every renderer */
  {
    FILE *file = g_fopen(filename, "w");

    if (!file) {
      message_error(_("Can't open output file %s: %s\n"), 
		    dia_message_filename(filename), strerror(errno));
      return;
    }
    fclose(file);
  }
  renderer = DRS_RENDERER (g_object_new(DRS_TYPE_RENDERER, NULL));
  /* store also object properties */
  renderer->save_props = (user_data == NULL);

  /* set up the root node */
  doc = xmlNewDoc((const xmlChar *)"1.0");
  doc->encoding = xmlStrdup((const xmlChar *)"UTF-8");
  doc->standalone = FALSE;
  dtd = xmlCreateIntSubset(doc, (const xmlChar *)"drs",
		     (const xmlChar *)"-//DIA//DTD DRS " VERSION "//EN",
		     (const xmlChar *)"http://projects.gnome.org/dia/dia-render-script.dtd");
  xmlAddChild((xmlNodePtr) doc, (xmlNodePtr) dtd);
  renderer->root = xmlNewDocNode(doc, NULL, (const xmlChar *)"drs", NULL);
  xmlAddSibling(doc->children, (xmlNodePtr) renderer->root);
  
  drs_data_render(data, DIA_RENDERER(renderer));

  xmlSetDocCompressMode(doc, 1);
  xmlDiaSaveFile(filename, doc);
  xmlFreeDoc(doc);

  g_object_unref(renderer);
}

static const gchar *extensions[] = { "drs", NULL };
DiaExportFilter export_filter = {
  N_("DiaRenderScript"),
  extensions,
  export_data
};

DIA_PLUGIN_CHECK_INIT

PluginInitResult
dia_plugin_init(PluginInfo *info)
{
  if (!dia_plugin_info_init(info, "drs",
			    N_("DiaRenderScript filter"),
			    NULL, NULL))
    return DIA_PLUGIN_INIT_ERROR;

  filter_register_export(&export_filter);

  return DIA_PLUGIN_INIT_OK;
}
